<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>{% block title%}{% endblock %}</title>
    <!-- Required meta tags -->
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
    <link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css" integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T" crossorigin="anonymous">
    <style>
        html {
            font: 100%/1.5 "Open Sans", sans-serif;
        }

        body {
            background-color: #eee;
        }

        h1, h2, h3, h4, h5, h6 {
            margin: 0;
        }

        a {
            color: #45b;
            text-decoration: none;
        }

        a:hover {
            text-decoration: underline;
        }

        input {
            line-height: 1.5 !important;
        }

        pre, code {
            font-family: "Menlo", "Monaco", "Consolas", monospace;
        }

        .button {
            background-color: #69b;
            box-shadow: 0 0.3333rem #479;
            color: #fff;
            transition: opacity 150ms ease-out,
                        transform 20ms ease-out,
                        box-shadow 20ms ease-out;
            border: none;
            border-radius: 0.3333rem;
            padding: 0.25em 1em;
        }

        .button:focus {
            outline: none;
            box-shadow: 0 0 0.0833rem 0.1667rem rgba(128,128,255,0.5), 0 0.3333rem #479;
        }

        .button:active:not(:disabled) {
            transform: translateY(0.3333rem);
            box-shadow: 0 0 #479;
        }

        .button:disabled {
            background-color: #9ab;
            opacity: 0.5;
            box-shadow: 0 0 #479;
        }

        #header {
            background-color: #000055;
            text-align: center;
        }

        #content {
            padding-top: 20px;
        }

    </style>
    {% block cs %}{% endblock %}

    <!-- jQuery first, then Popper.js, then Bootstrap JS -->
    <script src="/js/jquery-3.5.0.min.js"></script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.7/umd/popper.min.js" integrity="sha384-UO2eT0CpHqdSJQ6hJty5KVphtPhzWj9WO1clHTMGa3JDZwrnQq4sF86dIHNDz0W1" crossorigin="anonymous"></script>
    <script src="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/js/bootstrap.min.js" integrity="sha384-JjSmVgyd0p3pXB1rRibZUAYoIIy6OrQ6VrjIEaFf/nJGzIxFDsf4x0xIM+B07jRM" crossorigin="anonymous"></script>
    <!-- Bootstrap CSS -->

    {% block js %}{% endblock %}
</head>
<body>
    <!--
    <div id="top_menu">
        <ul>
            <li><a href="/admin/list_certs.html">List x509 Certificates</a></li>
            <li><a href="/admin/pending_csrs.html">Pending CSR Requests</a></li>
            <li><a href="/admin/create_csr.html">Create CSR Request</a></li>
        </ul>
    </div>-->
    <div class="container">
        <div id="header"><img height="100" id="logotag" src="/logo-tagline.png" /></div>
        {% block content %}{% endblock %}
    </div>

</body>
</html>
